' Enigma Cipher Machine emulation for CMM2
' Data taken from "Inside Enigma" by Prof. TomPerera, W1TP,
' Radio Society of Great Britian, plus private correspondence.
' Rev 1.0.0 William M Leue 1/3/2021

option default integer
option base 1

' Fundamental Machine Constants
const NUM_MOVING_WHEELS = 5
const NUM_REFLECTORS = 2
const NUM_WHEELS = NUM_MOVING_WHEELS+NUM_REFLECTORS
const NUM_CHARS  = 26
const TO_REFLECTOR = 1
const FROM_REFLECTOR = -1
const NUM_WHEELS_USED = 4
const MAX_PBJUMPERS = 13
const NUM_ROWS = 3
const NUM_COLS = 9  ' includes a dummy for middle row kb and lites

' Main Menu
const MENUX = 150
const MENUY = 100
const MENUW = 500
const MENUH = 40

const MHELP = 1
const MSETP = 2
const MUSE  = 3
const MQUIT = 4
const NMENUS = 4
const NHELP_PAGES = 10

' Machine Top Display constants
const MWIDTH = 400
const MHEIGHT = 500
const S1HEIGHT = 150
const S2HEIGHT = 175
const S3HEIGHT = 175
const KLRAD = 10
const KLHSEP = 40
const KLVSEP = 45
const NROWS = 3
const BWIDTH = 5
const WHEIGHT = 90
const WWIDTH = 8
const NWS = 10
const WSEP  = 50
const AWIDTH = 20
const AHEIGHT = 20
const BCOLOR = RGB(130, 30, 30)
const MCOLOR = RGB(32, 32, 32)

' Machine Setup Display constants
const S4HEIGHT = 300
const RSEP = 50
const RBRAD = 25
const RBCNT = 4
const SETUPX = 50
const SETUPTX = 455
const RBHSEP = 3*RBRAD
const RBVSEP = 4*RBRAD
const PBVSEP = 55
const PBRAD = 4
const PBYS2 = 4
const RBRAD2 = 29
const RBANG = 13.846
const RBRAD3 = 3
const RBARC1 = 170.0
const RBARC2 = 260.0
const RHEIGHT = 2*RBRAD2
const RWIDTH =  30

' text window constants
const TXX = MWIDTH+30
const TXY = MM.VRES\2 - MHEIGHT\2
const TXW = MM.HRES-TXX-1
const TXH = MHEIGHT-100
const TXD = TXH\2
const TXT_MAX_CCLINE = 40
const TXT_MAX_LINES = 12
const NCGROUP = 5

' keyboard commands (ASCII)
const UP     = 128
const DOWN   = 129
const LEFT   = 130
const RIGHT  = 131
const ENTER  = 13
const DELETE = 127
const ESCAPE = 27
const SPACE  = 32
const BKSPAC = 8
const HOME   = 134
const KEND   = 135
const PAGUP  = 136
const PAGDN  = 137

' Main Screen modes
const PREP = 1
const ENCMODE = 2
const DECMODE = 3

' Globals
dim menu_last_choice = 1
dim wheels(NUM_WHEELS, 2, NUM_CHARS)
dim notches(NUM_MOVING_WHEELS)
dim turnovers(NUM_MOVING_WHEELS)
dim keyLabels$(3, 9)
dim wheelLabels$(NUM_WHEELS) = ("I", "II", "III", "IV", "V", "B", "C")
dim letter_order(3, 9)

' Tables of Settings Used for an Encryption/Decryption
' Wheels are listed right to left, and the 
' left-most wheel must be a reflector. The
' encryption path is:
' Key Pressed -> Plugboard -> Rotor1 -> Rotor2 -> Rotor3 ->
' Reflector -> Rotor3 -> Rotor2 -> Rotor1 -> Plugboard -> Light
dim wheels_used(NUM_WHEELS_USED)
dim index_ring_settings(NUM_MOVING_WHEELS)
dim starting_wheel_positions(NUM_WHEELS)
dim plugboard_used(NUM_CHARS)
dim num_plugs_used = 0
dim plug_colors(NUM_CHARS)

' Current wheel positions and notch engagement
dim wheel_positions(NUM_WHEELS)
dim notch_engaged(NUM_WHEELS, 2)
dim rotor_selected = 0

' Text Windows
dim cipher_text_nlines = 0
dim plain_text_nlines = 0
dim cipher_text_lines$(TXT_MAX_LINES)
dim plain_text_lines$(TXT_MAX_LINES)
dim plain_text_numchars = 0
dim cipher_text_linesb$(TXT_MAX_LINES)
dim cipher_text_nlinesb = 0
dim total_num_cipher_chars = 0
dim mmode = ENCMODE

' Help System
dim pnames$(NHELP_PAGES)
dim last_help_page = 1

' Main Program
open "debug.txt" for output as #1
SetGraphics
ReadCodeWheels
VerifyCodeWheels
ReadKeys
MakePlugColors
LoadSetup
HandleMenuEvents
ClearTextWindow
close #1
end

' Set up Graphics
sub SetGraphics
  mode 1,8
end sub

' Read the data for the wiring nests, notches, and turnovers
' for all 8 rotating wheels and 2 reflector wheels.
sub ReadCodeWheels
  local i, j, d, nd
  for i = 1 to NUM_WHEELS
    nd = 2
    if i > 5 then
      nd = 1
    end if
    for d = 1 to nd
      for j = 1 to NUM_CHARS
        read wheels(i, d, j)
      next j
    next d
  next i
  for i = 1 to NUM_MOVING_WHEELS
    read notches(i)
  next i
  for i = 1 to NUM_MOVING_WHEELS
    read turnovers(i)
  next i
end sub

' Verify the wheel data: make sure each wheel has all the
' letter positions 1..26. If the wheel has both forward and
' reverse nest mappings, check both directions and make sure
' they are consistent.
sub VerifyCodeWheels
  local i, j, c, ok, n, d, nd
  local slots(2, NUM_CHARS)
  for i = 1 to NUM_WHEELS
    nd = 2
    if i > 5 then
      nd = 1
    end if
    for d = 1 to nd
      for j = 1 to NUM_CHARS
        slots(d, j) = 0
      next j
    next d
    for d = 1 to nd
      for j = 1 to NUM_CHARS
        c = wheels(i, d, j)
          if c < 1 or c > NUM_CHARS then
            print #1, "Illegal value in data for wheel " + str$(i);
            print #1, " at position (" + str$(d) + "," + str$(j) + ")"
          else
            slots(d, c) = 1
          end if
      next j
    next d
    if nd = 2 then
      for j = 1 to NUM_CHARS
        if wheels(i, 1, j) = c then
          if wheels(i, 2, c) <> j then
            n = n+1
            ok = 0
          end if
        end if
      next j
    end if
    ok = 1
    n = 0
    for d = 1 to nd
      for j = 1 to NUM_CHARS
        if slots(d, j) = 0 then
          ok = 0
          n = n+1
        end if
      next j
    next d
    if not ok then
      print #1, "Wheel data for wheel " + str$(i) + " has " + str$(n) + " defects"
    else
      print #1, "The wheel data for wheel " + str$(i) + " has no defects"
    end if
  next i
end sub

' Read the keyboard and lights labels in raster order
sub ReadKeys
  local row, col
  for row = 1 to 3
    for col = 1 to 9
      read keylabels$(row, col)
    next col
  next row
  for row = 1 to 3
    for col = 1 to 9
      read letter_order(row, col)
    next col
  next row
end sub

' Make a rainbow of colors for the plugboard plugs and wires
sub MakePlugColors
  local i, r, g, b
  local float h, s, v
  s = 0.7
  v = 1.0
  for i = 1 to NUM_CHARS
    h = (i-1)*13.85
    HSV2RGB h, s, v, r, g, b
    plug_colors(i) = RGB(r, g, b)
  next i
end sub

' Show the Menu of available commands
sub DrawMenuScreen
  local x, y, i

  cls
  x = MENUX
  y = MENUY
  text MM.HRES\2, 20, "Enigma Cipher Machine", "CT", 5
  text MM.HRES\2, 60, "Use up and down arrows to navigate, Enter to execute", "CT"
  InitHelp
  for i = 1 to NMENUS
    box x, y, MENUW, MENUH
    select case i
      case MHELP
        text x+10, y+15, "Help with Enigma and this Program"
      case MSETP
        text x+10, y+15, "Edit Machine Setup"
      case MUSE
        text x+10, y+15, "Use Enigma"
      case MQUIT
        text x+10, y+15, "Quit"
    end select
    y = y + MENUH+10
  next i
  HiliteMenu menu_last_choice
end sub

sub HandleMenuEvents
  local cmd
  local z$
  DrawMenuScreen
  choice = menu_last_choice
  do
    z$ = INKEY$
    do
      z$ = INKEY$
    loop until z$ <> ""
    cmd = asc(z$)
    select case cmd
      case UP
        choice = choice-1
        if choice < 1 then choice = 1
        HiliteMenu choice
      case DOWN
        choice = choice+1
        if choice > NMENUS then choice = NMENUS
        HiliteMenu choice
      case ENTER
        menu_last_choice = choice
        select case choice
            case MHELP
              HandleHelpEvents last_help_page
              DrawMenuScreen
            case MSETP
              SetUpEventLoop
              DrawMenuScreen
            case MUSE
              MachineEventLoop
              DrawMenuScreen
            case MQUIT
              close #1
              cls
              end
        end select
    end select
  loop  
end sub

' Hilite the current Menu Line
sub HiliteMenu which
  local xv(4), yv(4)
  local c

  if prev_menu > 0 then
    xv(1) = MENUX-10        : yv(1) = MENUY+(prev_menu-1)*(MENUH+10)+MENUH\2
    xv(2) = xv(1)-20        : yv(2) = yv(1) - MENUH\2
    xv(3) = xv(2)           : yv(3) = yv(1) + MENUH\2
    xv(4) = xv(1)           : yv(4) = yv(1)
    polygon 4, xv(), yv(),  RGB(BLACK), RGB(BLACK)
  end if
  xv(1) = MENUX-10        : yv(1) = MENUY+(which-1)*(MENUH+10)+MENUH\2
  xv(2) = xv(1)-20        : yv(2) = yv(1) - MENUH\2
  xv(3) = xv(2)           : yv(3) = yv(1) + MENUH\2
  xv(4) = xv(1)           : yv(4) = yv(1)
  polygon 4, xv(), yv(),  RGB(RED), RGB(RED)
  prev_menu = which
end sub


' perform correct wraparound into alphabet range
' This is used to handle any calculations involving
' input and output wires, wheel positions, and
' index ring positions so that numerical wraps from
' the ends of the alphabet are done correctly.
function wrap(k)
  local w
  if k >= 1 and k <= NUM_CHARS then
    w = k
  else if k > 26 then
    w = k mod NUM_CHARS
  else if k <= 0 then
    w = NUM_CHARS + k
  end if
  wrap = w
end function

' given the input wire, rotor position, and ring setting,
' return the rotor nest input index
function NestInput(iwire, rpos, ring)
  NestInput = wrap(iwire + (rpos-1) - (ring-1))
end function

' given a rotor number, nest input index, return the nest output index,
' depending on the signal direction.
function NestOutput(rotor, inest, dir)
  if dir > 0 then
    NestOutput = wheels(rotor, 1, inest)
  else
    NestOutput = wheels(rotor, 2, inest)
  end if
end function

' given a nest output index, rotor position, and ring setting,
' return the output wire index.
function OutputWire(onest, rpos, ring)
  OutputWire = wrap(onest - rpos + ring)
end function

' Compute the output wire number (1..26) for a given wheel.
' wheel: wheel number (I..VII = 1..8)
' inwire: index of input wire 1..26
' direction: 1 = forward toward reflector, -1 = away from reflector
function GetWheelOutput(wheel, inwire, direction)
  local rpos, ring, inest, onest, outwire
  rpos = wheel_positions(wheel)
  ring = index_ring_settings(wheel)
  inest = NestInput(inwire, rpos, ring)
  onest = NestOutput(wheel, inest, direction)
  outwire = OutputWire(onest, rpos, ring)
  GetWheelOutput = outwire
end function

' Given the input wire, return the output wire for the selected
' reflector. This is greatly simplified compared to a moving
' rotor because the reflectors cannot move, have no index ring,
' and the signals only move in one direction.
function GetReflectorOutput(wheel, inwire)
  local outwire
  outwire = NestOutput(wheel, inwire, 1)
  GetReflectorOutput = outwire
end function
  
' Set up some simple initial conditions:
' 1. Use rotors I, II, III, and B right to left
' 2. Index Rings all set to the 'A' position.
' 3. Moving wheels all set to 'A' position.
' 4. No plugs in plugboard.
' The user can change any and all of these settings.
' Only called if there is no saved setup.
sub InitialSetup
  local i
  wheels_used(1) = 1
  wheels_used(2) = 2
  wheels_used(3) = 3
  wheels_used(4) = 6
  for i = 1 to NUM_MOVING_WHEELS
    index_ring_settings(i) = 1
    starting_wheel_positions(i) = 1
    wheel_positions(i) = 1
  next i
  for i = 1 to MAX_PBPUMPERS
    plugboard_used(i) = 0
    num_plugs_used = 0
  next i
end sub

' Get User inputs from keyboard and/or Classic Controller for Machine setup
' Can be done using keyboard or Classic Controller.
' Things that can be changed:
'  1. Rotors used and their position on shaft
'  2. Reflectors used
'  3. Index ring position on moving rotors
'  4. Plugboard connections.
sub SetupEventLoop
  local z$
  local cmd, brn, bdest, npsrc, pdst, npos, dup, remove_plug, nstate
  local state = 0
  DrawSetup
  z$ = INKEY$
  do
    do
      z$ = INKEY$
    loop until z$ <> ""
    cmd = asc(UCASE$(z$))
    if cmd = ESCAPE then
      SaveSetup
      exit do
    end if
    select case state
      case 0
        if UCASE$(z$)      = "R" then
          nstate = 1
        else if UCASE$(z$) = "F" then
          nstate = 3
        else if UCASE$(z$) = "P" then
          nstate = 4
        else if UCASE$(z$) = "I" then
          nstate = 6    
        else
          Beep
          nstate = 0
        end if
print #1, "state 0: nstate = " + str$(nstate)
      case 1
        brn = cmd - asc("1") + 1
print #1, "state 1: cnd: " + str$(cmd) + " brn: " + str$(brn)
        if brn < 1 or brn > NUM_MOVING_WHEELS then
          ShowError state
          brn = -1
        else
          for j = 1 to NUM_WHEELS_USED
            if wheels_used(j) = brn then
              brn = -1
            end if
          next j
          if brn < 0 then            
            ShowError state
          else
            nstate = 2
            rotor_selected = brn
            DrawSetup
          end if
        end if
print #1, "nstate: " + str$(nstate)
      case 2
        bdest = -1
        cmd = asc(UCASE$(z$))
        bdest = cmd - asc("1") + 1
print #1, "state 2: bdest: " + str$(bdest)
        if bdest < 1 or bdest > 3 then
          ShowError state
          bdest = -1
        else
          nstate = 0
print #1, "calling MountWheel"
          MountWheel brn, bdest
        end if
      case 3
        brn = -1
        cmd = asc(UCASE$(z$))
        brn = cmd - asc("A") + 5
        if brn < 6 or brn > 7 then
          ShowError state
          brn = -1
        else
          nstate = 0
          MountWheel brn, 4
        end if
      case 4
        remove_plug = 0
        delete_all = 0
        psrc = -1
        if z$ = "-" then
          remove_plug = 1
          nstate = 5
        else
          cmd = asc(UCASE$(z$))
          if cmd = DELETE then
            for i = 1 to NUM_CHARS
              plugboard_used(i) = 0
            next i
            num_plugs_used = 0
            nstate = 0
            DrawSetup
          else
            psrc = cmd - asc("A") + 1
            if psrc < 1 or psrc > NUM_CHARS then
              ShowError state
              psrc = -1
            else if num_plugs_used = MAX_PBJUMPERS then
              ShowError state
            else
              dup = 0
              if plugboard_used(psrc) > 0 then
                dup = 1
                ShowError state
              end if
              if dup = 0 then
                nstate = 5
              end if
            end if
          end if
        end if
      case 5
        pdst = -1
        cmd = asc(UCASE$(z$))
        pdst = cmd - asc("A") + 1
        if pdst < 1 or pdst > NUM_CHARS or psrc = pdst then
          ShowError state
          pdst = -1
        else
          if remove_plug then
            hit = 0
            if plugboard_used(pdst) > 0 then
              posrc = plugboard_used(pdst)
              plugboard_used(pdst) = 0
              plugboard_used(posrc) = 0
              num_plugs_used = num_plugs_used-1
              hit = 1
            end if
            remove_plug = 0
            nstate = 0
            DrawSetup
          else
            dup = 0
            if plugboard_used(pdst) > 0 then
              dup = 1
            end if
            for i = 1 to NUM_CHARS
              if plugboard_used(i) = pdst then
                dup = 1
                exit for
              end if
            next i
            if dup = 1 then
              ShowError state
            else          
              nstate = 0
              MountPlug psrc,pdst
            end if
          end if
        end if
      case 6
        brn = -1
        cmd = asc(UCASE$(z$))
        brn = cmd - asc("1") + 1
        if brn < 1 or brn > NUM_MOVING_WHEELS then
          ShowError state
          brn = -1
        else
          rotor_selected = brn
          DrawSetup
          nstate = 7
        end if
      case 7
        cmd = asc(UCASE$(z$))
        if cmd < asc("A") or cmd > asc("Z") then
          ShowError state
        else
          index_ring_settings(brn) = cmd - asc("A") + 1
          nstate = 0
          rotor_selected = 0
          DrawSetup
        end if
    end select
    state = nstate
  loop
end sub

' Mount a Wheel in the specified position
' If a wheel is already in that position, it will
' be replaced by the specified wheel.
sub MountWheel brd, bdest
  wheels_used(bdest) = brd
  DrawSetup
end sub

' Add a plug cable to the Plugboard
' (Plug removes are handled in the event handler)
sub MountPlug psrc,pdst
  if plugboard_used(psrc) = 0 then
    num_plugs_used = num_plugs_used+1
  end if
  plugboard_used(psrc) = pdst
  plugboard_used(pdst) = psrc
  DrawSetup
end sub

' Show a User Error in Setup
sub ShowError state
  Beep 
  text 50, 550, "Error - Try Again"
  pause 1000
  text 50, 550, "                 "
  state = 0
  rotor_selected = 0
end sub

sub Beep 
  play sound 1, B, T, 800
  pause 250
  play stop
end sub

' Encrypt or Decrypt a Single Character, based on the curremt machine setup
function EncryptWire(inwire)
  AdvanceWheels
  local pbout1, pbout2, tchar, inum, onum
  local outchar

  pbout1 = PlugBoardOutput(inwire, TO_REFLECTOR)
  tchar = pbout1
  for i = 1 to NUM_WHEELS_USED-1
    tchar = GetWheelOutput(wheels_used(i), tchar, TO_REFLECTOR)
  next i
  tchar = GetReflectorOutput(wheels_used(4), tchar)
  for i = NUM_WHEELS_USED-1 to 1 step -1
    tchar = GetWheelOutput(wheels_used(i), tchar, FROM_REFLECTOR)
  next i
  pbout2 = PlugBoardOutput(tchar, FROM_REFLECTOR)
  outwire = pbout2
  EncryptWire = outwire
end function

' Called each time a key is pressed. Advances any wheels that are turning:
' Rightmost wheel: every key press
' Next wheel: every time the rightmost wheel advances when its notch is engaged.
' Next wheel: every time the center wheel advances when its notch is engaged.
sub AdvanceWheels
  local notch1, notch2, w1, w2, w3
  
  w1 = wheels_used(1) : w2 = wheels_used(2) : w3 = wheels_used(3)

  wheel_positions(w1) = wheel_positions(w1)+1
  if wheel_positions(w1) > NUM_CHARS then wheel_positions(w1) = 1

  ' advancing wheel in position 2 when notch 1 is engaged
  if wheel_positions(w1) = turnovers(w1) and notch_engaged(w1, 2) = 0 then
    notch_engaged(w1, 1) = 1
  end if
  if notch_engaged(w1, 1) and notch_engaged(w1, 2) = 0 then
    wheel_positions(w2) = wheel_positions(w2)+1
    if wheel_positions(w2) > NUM_CHARS then wheel_positions(w2) = 1
    notch_engaged(w1, 1) = 0 : notch_engaged(w1, 2) = 1
  end if
  if wheel_positions(w1) <> turnovers(w1) then
    notch_engaged(w1, 2) = 0
  end if

  ' advancing wheel in position 3 when notch 2 is engaged
  if wheel_positions(w2) = turnovers(w2) and notch_engaged(w2, 2) = 0 then
    notch_engaged(w2, 1) = 1
  end if
  if notch_engaged(w2, 1) and notch_engaged(w2, 2) = 0 then  
    wheel_positions(w3) = wheel_positions(w3)+1
    if wheel_positions(w3) > NUM_CHARS then wheel_positions(w3) = 1
    notch_engaged(w2, 1) = 0 : notch_engaged(w2, 2) = 1
  end if
  if wheel_positions(w2) <> turnovers(w2) then
    notch_engaged(w2, 2) = 0
  end if
  DrawMachine
end sub

' Rewind the movable rotors 1 click. This will turn the right-most 
' rotor the requested (positive) number, and also any of the
' other wheels that were advanced by the previous 'n' clicks,
sub RewindWheels
  local notch1, notch2, w1, w2, w3, t1, t2, bt1, bt2
  
  w1 = wheels_used(1) : w2 = wheels_used(2) : w3 = wheels_used(3)

  ' rewind the rightmost wheel one step
  wheel_positions(w1) = wheel_positions(w1)-1
  if wheel_positions(w1) < 1 then wheel_positions(w1) = NUM_CHARS
  
  ' rewind wheel in position 2 one step when notch wheel 1 is at the turnover.
  ' If necessary, rewind wheel 3 a step also.
  bt1 = turnovers(w1)-1
  if bt1 < 1 then bt1 = NUM_CHARS
  if wheel_positions(w1) = bt1 then
    wheel_positions(w2) = wheel_positions(w2)-1
    if wheel_positions(w2) < 1 then wheel_positions(w2) = NUM_CHARS
    bt2 = turnovers(w2)-1
    if bt2 < 1 then bt2 = NUM_CHARS
    if wheel_positions(w2) = bt2 then
      wheel_positions(w3) = wheel_positions(w3)-1
      if wheel_positions(w3) < 1 then wheel_positions(w3) = NUM_CHARS
    end if
  end if
end sub

' Peform the plugboad scrambling for a character if there
' is a plug there.
function PlugboardOutput(inchar, direction)
  local outchar
  if plugboard_used(inchar) > 0 then
    outchar = plugboard_used(inchar)
  else
    outchar = inchar
  end if
  PlugboardOutput = outchar
end function

' Draw the Enigma Machine as seen from overhead
' key_down and light_on are key pressed and
' lamp illuminated locations, encoded as
' row*100 + col.
sub DrawMachine key_down, light_on
  local x, y
  page write 1
  box 0, 0, MWIDTH, MHEIGHT,, RGB(BLACK), RGB(BLACK) 
  y = MM.VRES\2 - MHEIGHT\2
  x = 10
  box x, y, MWIDTH, MHEIGHT, BWIDTH, MCOLOR, MCOLOR
  box x, y, MWIDTH, MHEIGHT, BWIDTH, BCOLOR
  line x+BWIDTH, y+S1HEIGHT, x+MWIDTH-2*BWIDTH+4, y+S1HEIGHT
  line x+BWIDTH, y+S1HEIGHT+S2HEIGHT, x+MWIDTH-2*BWIDTH+4, y+S1HEIGHT+S2HEIGHT
  DrawTopSection x,y
  if key_down > 0 then
    DrawLightSection x, y+S1HEIGHT, light_on
    DrawKeyboardSection x, y+S1HEIGHT+S2HEIGHT, key_down
    page write 0
    page copy 1 to 0,b
    pause 250
  end if
  page write 1
  DrawLightSection x, y+S1HEIGHT
  DrawKeyboardSection x, y+S1HEIGHT+S2HEIGHT
  DrawTextWindow
  UpdateTextWindow
  page write 0
  page copy 1 to 0,b
  'save image "machine"
end sub

' Draw the top section with the wheels and alphabet windows
sub DrawTopSection x, y
  local xc, yc, wx, wy, ax, ay, sy, i, j, c, wn
  xc = x + MWIDTH\2
  yc = y + S1HEIGHT\2
  wx = xc - WSEP
  wy = yc - WHEIGHT\2
  for i = 1 to NUM_WHEELS_USED-1
    wn = 4-i
    c = RGB(WHITE)
    if wn = rotor_selected then
      c = RGB(GREEN)
    end if
    box wx, wy, WWIDTH, WHEIGHT,, c, c
    for j = 1 to NWS
      sy = wy + j*WHEIGHT\NWS
      line wx, sy, wx+WWIDTH, sy, 3, RGB(BLACK)
    next j
    ax = wx - AWIDTH
    ay = yc - AHEIGHT\2
    box ax, ay, AWIDTH, AHEIGHT,, RGB(BLACK), RGB(BLACK)
    text ax+4, ay+4, chr$(wheel_positions(wn)+asc("A")-1))
    wx = wx + WSEP    
  next i
end sub

' Draw the section with the alphabet lights
' If light_on is not zero, draw the light illuminated.
sub DrawLightSection x, y, light_on
  local lx, ly, row, col, nc, rx, lrow, lcol
  lrow = light_on\100
  lcol = light_on - lrow*100
  for row = 1 to 3
    ly = y + row*KLVSEP
    nc = NUM_COLS
    if row = 2 then nc = NUM_COLS-1
    rx = x
    if row = 2 then rx = x + KLHSEP\2
    for col = 1 to nc
      lx = rx + col*KLHSEP
      if row = lrow and col = lcol then
        circle lx, ly, KLRAD,,, RGB(BLACK), RGB(YELLOW)
        text lx+1, ly+1, keyLabels$(row, col), "CM", 4,, RGB(BLACK), RGB(YELLOW)
      else
        circle lx, ly, KLRAD,,, RGB(BLACK), RGB(BLACK)
        text lx+1, ly+1, keyLabels$(row, col), "CM", 4,, RGB(WHITE), RGB(BLACK)
      end if
    next col
  next row
end sub

' Draw the section with the keyboard
' If key_down is not zero, draw the key with reversed colors
sub DrawKeyboardSection x, y, key_down
  local lx, ly, row, col, nc, rx, krow, kcol
  krow = key_down\100
  kcol = key_down - krow*100

  for row = 1 to 3
    ly = y + row*KLVSEP
    nc = NUM_COLS
    if row = 2 then nc = NUM_COLS-1
    rx = x
    if row = 2 then rx = x + KLHSEP\2
    for col = 1 to nc
      lx = rx + col*KLHSEP
      if row = krow and col = kcol then
        circle lx, ly, KLRAD,,, RGB(BLACK), RGB(WHITE)
        text lx+1, ly+1, keyLabels$(row, col), "CM", 4,, RGB(BLACK), RGB(WHITE)
      else
        circle lx, ly, KLRAD,,, RGB(WHITE), RGB(BLACK)
        text lx+1, ly+1, keyLabels$(row, col), "CM", 4,, RGB(WHITE), RGB(BLACK)
      end if
    next col
  next row
end sub

' Event loop for Machine Top View
' Handles rotor position setting and encryption/decryption
sub MachineEventLoop
  local z$, echar$, pchar$, eline$
  local cmd, i, anum, ncline

  z$ = INKEY$
  rotor_selected = 1
  for i = 1 to NUM_WHEELS_USED-1
    wheel_positions(i) = 1
  next i
  page write 1
  cls
  DrawMachine
  ClearTextWindow
  DrawTextWindow
  page write 0
  page copy 1 to 0
  do
    do
      z$ = INKEY$
    loop until z$ <> ""
    cmd = asc(UCASE$(z$))
    select case cmd
      case UP
        wheel_positions(rotor_selected) = wheel_positions(rotor_selected)+1
        if wheel_positions(rotor_selected) > NUM_CHARS then
          wheel_positions(rotor_selected) = 1
        end if
        DrawMachine
      case DOWN
        wheel_positions(rotor_selected) = wheel_positions(rotor_selected)-1
        if wheel_positions(rotor_selected) < 1 then
          wheel_positions(rotor_selected) = NUM_CHARS
        end if
        DrawMachine
      case LEFT
        rotor_selected = rotor_selected+1
        if rotor_selected > 3 then rotor_selected = 1
        DrawMachine
      case RIGHT
        rotor_selected = rotor_selected-1
        if rotor_selected < 1 then rotor_selected = 3
        DrawMachine    
      case ESCAPE
        exit do    
      case BKSPAC
        RewindEncryption 1, 3
      case HOME
        RewindEncryption -1, 3
      case KEND
        ToggleEncDecMode
      case else
        if cmd >= asc("A") and cmd <= asc("Z") then
          PressKey cmd
        end if
    end select
  loop
end sub

' Press an Alphabetic Key on the Enigma Keyboard to
' cause an encryption or decryption to occur.
sub PressKey cmd
  local pchar$, echar$
  local inwire, outwire, key_down, light_on
  inwire = cmd - asc("A") + 1
  pchar$ = chr$(cmd)
  outwire = EncryptWire(inwire)
  echar$ = chr$(asc("A") + outwire - 1)
  key_down = FindKeyLightIndex(inwire)
  light_on = FindKeyLightIndex(outwire)
  UpdatePlainText pchar$
  UpdateCipherText echar$
  DrawMachine key_down, light_on
end sub

' Rewind the encryption the specifed number of
' characters. If more than the current number
' of encrypted characters is requested, it will
' rewind however many it can. Removes the characters
' from the Text Window and rewinds the rotors, including
' any that were stepped by their neighbor.
' If 'n' is < 0 then ALL text will be deleted and the
' rotors rewound to the starting positions.
' w: 1 = clear plaintext
'    2 = clear ciphertext
'    3 = clear both
sub RewindEncryption n, w
  local num_existing, i
  num_existing = plain_text_numchars
  if num_existing = 0 then
    exit sub
  end if
  
  if n < 0 then n = plain_text_numchars
  for i = 1 to n
    if w = 1 or w = 3 then
      UpdatePlainText chr$(BKSPAC)
    end if
    if w = 2 or w = 3 then
      UpdateCipherText chr$(BKSPAC)
      total_num_cipher_chars = total_num_cipher_chars-1
    end if
    RewindWheels
  next i
  DrawMachine
end sub

' Toggle between Encode Mode (ENCMODE) and Decode Mode (DECMODE)
sub ToggleEncDecMode
  if mmode = ENCMODE then
    mmode = DECMODE
  else
    mmode = ENCMODE
  end if
  DrawMachine
end sub

' Given an wire value, return the key or light
' row and column for it, encoded as row*100 + col
function FindKeyLightIndex(wire)
  local row, col, hit, nc
  hit = 0
  for row = 1 to NUM_ROWS
    nc = NUM_COLS
    if row = 2 then nc = NUM_COLS-1
    for col = 1 to nc
      if letter_order(row, col) = wire then
        hit = row*100+col
        exit for
      end if
    next col
    if hit > 0 then exit for
  next row
  FindKeyLightIndex = hit    
end function

' clear the character arrays in the text window
sub ClearTextWindow
  ClearPlaintextWindow
  ClearCiphertextWindow
end sub

sub ClearPlaintextWindow
  local i
  plain_text_nlines = 1
  for i = 1 to TXT_MAX_LINES
    plain_text_lines$(TXT_MAX_LINES) = ""
  next i
  plain_text_numchars = 0
end sub

sub ClearCiphertextWindow
  local i
  cipher_text_nlines = 1
  for i = 1 to TXT_MAX_LINES
    cipher_text_lines$(TXT_MAX_LINES) = ""
  next i
end sub

' Draw the Text Window
' mmode: Encmode, Decmode
sub DrawTextWindow
  local x, y
  x = TXX
  y = TXY
  box x, y, TXW, TXH,,, RGB(BLACK)
  line x, y+20, x+TXW, y+20
  if mmode = ENCMODE then
    text x+TXW\2, y+6, "Cipher Text", "CT"
  else
    text x+TXW\2, y+TXD+6, "Plain Text", "CT"
  end if
  line x, y+TXD, x+TXW, y+TXD
  if mmode = ENCMODE then
    text x+TXW\2, y+TXD+6, "Plain Text", "CT"
  else
    text x+TXW\2, y+6, "Cipher Text", "CT"
  end if
  line x, y+TXD+20, x+TXW, y+TXD+20

  x = TXX+5
  y = TXY+TXH+10
  text x, y, "Backspace: back up encryption 1 char"    : y = y+15
  text x, y, "Home: erase encryption and reset wheels" : y = y+15
  text x, y, "End: Toggle Encode/Decode"               : y = y+15
  text x, y, "Up and Down Arrows: turn selected wheel" : y = y+15
  text x, y, "Left and Right Arrows: select wheel"     : y = y+15
  text x, y, "Escape: Return to Main Menu"             : y = y+15
end sub

' Redraw the plain and cipher text windows.
sub UpdateTextWindow
  local x, y, i, c
  x = TXX+5
  y = TXY+TXD+35
  c = RGB(CYAN)
  if mmode = DECMODE then
    c = RGB(RED)
  end if
  for i = 1 to plain_text_nlines
    text x, y, space$(TXT_MAX_CCLINE)
    text x, y, plain_text_lines$(i),,,, c
    y = y+15
  next i
  if plain_text_nlines <= 0 then
    plain_text_nlines = 1
  end if
  y = TXY+20+15
  c = RGB(RED)
  if mmode = DECMODE then
    c = RGB(CYAN)
  end if
  for i = 1 to cipher_text_nlines
    text x, y, space$(TXT_MAX_CCLINE)
    text x, y, cipher_text_lines$(i),,,, c
    y = y+15
  next i
  if cipher_text_nlines <= 0 then
    cipher_text_nlines = 1
  end if
end sub

' Update the Plain Text arrays when a new
' character comes in. If the new char is
' BACKSPACE, a char is removed.
sub UpdatePlainText pchar$
  local tlen, x, y
  local temp$
  if asc(pchar$) = BKSPAC then
    inc plain_text_numchars, -1
    temp$ = plain_text_lines$(plain_text_nlines)
    tlen = len(temp$)
    temp$ = LEFT$(temp$, tlen-1)
    plain_text_lines$(plain_text_nlines) = temp$
    if tlen = 1 then
      inc plain_text_nlines, -1
    end if
  else
    inc plain_text_numchars
    tlen = len(plain_text_lines$(plain_text_nlines))
    if tlen >= TXT_MAX_CCLINE then  
      inc plain_text_nlines
      if plain_text_nlines >= TXT_MAX_LINES then
        Beep
        exit sub
      end if
    end if
    cat plain_text_lines$(plain_text_nlines), pchar$
  end if
end sub

' Update the Cipher Text arrays when a new
' character comes in
sub UpdateCipherText pchar$
  local tlen, x, y, asp, nd
  local temp$
  if asc(pchar$) = BKSPAC then
    temp$ = cipher_text_lines$(cipher_text_nlines)
    tlen = len(temp$)
    nd = 1
    if RIGHT$(temp$, 1) = " " then nd = 2
    temp$ = LEFT$(temp$, tlen-nd)
    cipher_text_lines$(cipher_text_nlines) = temp$
    if len(cipher_text_lines$(cipher_text_nlines)) = 0 then
      inc cipher_text_nlines, -1
    end if
  else
    tlen = len(cipher_text_lines$(cipher_text_nlines))
    if tlen >= TXT_MAX_CCLINE then  
      inc cipher_text_nlines
      if cipher_text_nlines >= TXT_MAX_LINES then
        Beep
        exit sub
      end if
    end if
    cat cipher_text_lines$(cipher_text_nlines), pchar$
    tlen = tlen+1
    if mmode = ENCMODE then
      if (tlen+1) mod NCGROUP = 0 then
        cat cipher_text_lines$(cipher_text_nlines), " "
      end if
    end if
  end if
end sub

' Draw the setup parts
sub DrawSetup
  local x, y
  page write 1
  cls
  x = SETUPX
  y = MM.VRES\2 - MHEIGHT\2
  box x, y, MWIDTH, MHEIGHT, BWIDTH, MCOLOR, MCOLOR
  text x+MWIDTH-8, y+6, "Rotor Setup", "RT",,,, MCOLOR
  box x, y, MWIDTH, MHEIGHT, BWIDTH, BCOLOR
  line x+BWIDTH, y+S4HEIGHT, x+MWIDTH-2*BWIDTH+4, y+S4HEIGHT
  text x+MWIDTH-8, y+S4HEIGHT+6, "Plugboard Setup", "RT",,,, MCOLOR
  DrawRotorSetup x, y
  DrawPlugboardSetup x, y+S4HEIGHT
  DrawSetupHelp
  page write 0
  page copy 1 to 0, b
  save image "setup"
end sub

' Draw the Parts for the Rotor Setup
sub DrawRotorSetup x, y
  local i, j, rx, ry, row, col, bx, by, s, rbnum, cc, used
  local float rcang, rcarc1, rcarc2, rcx, rcy
  local irc$

  ry = y + 30
  box x + 60, ry + RHEIGHT\2 - 4, 230, 8
  for i = 1 to NUM_WHEELS_USED
    cc = RGB(WHITE)
    if i = 1 then
      cc = RGB(RED)
    end if
    rx = x + 80 + (i-1)*RSEP
    wnum = NUM_WHEELS_USED-i+1
    rh = RHEIGHT
    ro = 0
    if i = 1 then
      rh = RHEIGHT-10
      ro = 5
    end if
    box rx, ry+ro, RWIDTH, rh,, cc, RGB(WHITE)
    text rx+RWIDTH\2, ry-5, str$(wnum), "CB", 7,,, MCOLOR
    text rx+RWIDTH\2, ry+rh-(5-ro), wheelLabels$(wheels_used(wnum)), "CB", 7,, RGB(BLACK), RGB(WHITE)
  next i
  rbnum = 0
  for row = 1 to 2
    by = y + 20 + RHEIGHT + 60 + (row-1)*RBVSEP
    for col = 1 to RBCNT
      rbnum = rbnum+1
      if rbnum > NUM_WHEELS then exit for
      used = 0
      for j = 1 to NUM_WHEELS_USED
        if wheels_used(j) = rbnum then
          used = 1
          exit for
        end if
      next j
      bx = x + 80 + (col-1)*RBHSEP
      if used then
        circle bx, by, RBRAD2,,, RGB(75, 75, 75)
        text bx, by-9, wheelLabels$(rbnum), "CB", 7,, RGB(75, 75, 75), MCOLOR
        if rbnum <= NUM_MOVING_WHEELS then
          irc$ = chr$(asc("A") + index_ring_settings(rbnum)-1)
          text bx, by+RBRAD2+15, irc$, "CT", 7,, RGB(75,75,75), MCOLOR  
        end if
      else
        cc = RGB(WHITE)
        if rotor_selected = rbnum then
          cc = RGB(GREEN)
        else if rbnum > NUM_MOVING_WHEELS then
          cc = RGB(RED)
        end if
        circle bx, by, RBRAD,,, cc
        circle bx, by, 4,,, cc
        text bx, by-9, wheelLabels$(rbnum), "CB", 7,, cc, MCOLOR
        if rbnum <= NUM_MOVING_WHEELS then
          rcang = 0.0
          for s = 1 to NUM_CHARS
            rcx = bx + RBRAD2*cos(rad(rcang))
            rcy = by - RBRAD2*sin(rad(rcang))
            rcarc1 = RBARC1 - rcang
            rcarc2 = RVBARC2 - rcang
            arc rcx, rcy, RBRAD3,, rcarc1, rcarc2, cc
            rcang = rcang + RBANG
          next s
          irc$ = chr$(asc("A") + index_ring_settings(rbnum)-1)
          text bx, by+RBRAD2+15, irc$, "CT", 7,, RGB(WHITE), MCOLOR  
        end if
      end if
    next col
  next row
end sub

' Draw the Plugboard with any connected cables
sub DrawPlugboardSetup x, y
  local lx, ly, row, col, nc, rx, pnum
  pnum = 0
  for row = 1 to 3
    ly = y + (row-1)*PBVSEP + 40
    nc = 9
    if row = 2 then nc = 8
    rx = x
    if row = 2 then rx = x + KLHSEP\2
    for col = 1 to nc
      pnum = pnum+1
      lx = rx + col*KLHSEP
      circle lx, ly, PBRAD,,, RGB(BLACK), RGB(BLACK)
      circle lx, ly+PBYS2*PBRAD, PBRAD,,, RGB(BLACK), RGB(BLACK)
      text lx-PRAD-15, ly+PBRAD+3, keyLabels$(row, col), "CM", 4,, RGB(WHITE), MCOLOR
      if plugboard_used(pnum) > 0 then
        DrawPlugCable pnum, x, y
      end if
    next col
  next row  
end sub

' Draw a plugboard cable connecting two socket pairs
sub DrawPlugCable which, x, y
  local row, col, i, p, hit, c
  local lx(2), ly(2)

  c = plug_colors(which)
  for i = 1 to 2
    if i = 1 then
      p = which
    else 
      p = plugboard_used(which)
    end if
    hit = 0
    for row = 1 to NUM_ROWS
      for col = 1 to NUM_COLS
        if letter_order(row, col) = p then
          hit = 1
          exit for
        end if
      next col
      if hit = 1 then
        exit for
      end if
    next row  
    lx(i) = x + col*KLHSEP
    if row = 2 then lx(i) = lx(i) + KLHSEP\2
    ly(i) = y + (row-1)*PBVSEP + 40
    rbox lx(i)-7, ly(i)-7, 15, PBYS2+5*PBRAD+7, 3, c, c
  next i
  line lx(1), ly(1), lx(2), ly(2),, c
end sub

' Setup Enigma Help Text
sub DrawSetupHelp
  local x, y, c
  x = SETUPTX
  y = 50
  text x, y, "          Setup Commands"                 : y = y+15
  text x, y, "-----------------------------------------": y = y+15
  text x, y, "To choose rotors and their positions:"    : y = y+15
  text x, y, "  R <rotor_number 1-5> <shaft pos 1-3>"   : y = y+15
  text x, y, "  example:  R 5 2 puts rotor V in pos 2"  : y = y+15
  text x, y, "To choose rotor index ring positions:"    : y = y+15
  text x, y, "  I <rotor_number 1-5> <letter A-Z>"      : y = y+15
  text x, y, "  example:  I 4 R sets ring of IV to R"   : y = y+15
  text x, y, "You can set index rings of both used and" : y = y+15
  text x, y, "  unused rotors. Used rotors are dimmed." : y = y+15
  text x, y, "To choose Reflector B or C:"              : y = y+15
  text x, y, "  F <B or C>"                             : y = y+15
  text x, y, "  example: F C  chooses Reflector C."     : y = y+15
  text x, y, "To add or remove plugboard jumpers:"      : y = y+15
  text x, y, "  P <letter>  <letter>"                   : y = y+15
  text x, y, "  example: P Z A puts a jumper from Z-A"  : y = y+15
  text x, y, "To remove a jumper:"                      : y = y+15
  text x, y, "  P - <letter>"                           : y = y+15
  text x, y, "  example: P - Z removes jumper Z-A"      : y = y+15
  text x, y, "You can use either end to remove."        : y = y+15
  text x, y, "To remove ALL jumpers:"                   : y = y+15
  text x, y, "  P <delete>  This is faster."            : y = y+15
  text x, y, "Commands are not case-sensitive."         : y = y+15
  text x, y, ""                                         : y = y+15
  text x, y, "The setup does not include rotor start"   : y = y+15
  text x, y, "positions; that is on the next screen."   : y = y+15
  text x, y, "When your setup is complete, press the"   : y = y+15
  text x, y, "Escape key to return to the main menu."   : y = y+15
  y = y+15
  text x, y, "WARNING: responses to commands can be",,,, RGB(RED)
  y = y+15
  text x, y, "slow. Wait for visible change before",,,, RGB(RED)
  y = y+15
  text x, y, "entering the next command!",,,, RGB(RED)
end sub

' Save the Current Edited Enigma Setup
sub SaveSetup
  local i, n
  open "Enigma_Settings.sav" for output as #2
  print #2, "Enigma Settings"
  for i = 1 to NUM_WHEELS_USED
    print #2, str$(wheels_used(i)) + ",";
  next i
  print #2, ""
  for i = 1 to NUM_MOVING_WHEELS
    print #2, str$(index_ring_settings(i)) + ",";
  next i
  print #2, ""
  print #2, str$(num_plugs_used)
  n = 0
  for i = 1 to NUM_CHARS
    print #2, str$(plugboard_used(i)) + ",";
    inc n
    if n >= 10 then
      print #2, ""
      n = 0
    end if
  next i
  print #2, ""
  print #2, "end"
  close #2
end sub

' Load the Most Recent Enigma setup
sub LoadSetup
  local i, j
  local buf$, temp$
  on error skip 1
  open "Enigma_Settings.sav" for input as #2
  if MM.ERRNO <> 0 then
    InitialSetup
    SaveSetup
    exit sub
  end if
  line input #2, buf$
  line input #2, buf$
  for i = 1 to NUM_WHEELS_USED
    temp$ = FIELD$(buf$, i, ",")
    wheels_used(i) = val(temp$)
  next i
  line input #2, buf$
  for i = 1 to NUM_MOVING_WHEELS
    temp$ = FIELD$(buf$, i, ",")
    index_ring_settings(i) = val(temp$)
  next i
  line input #2, buf$
  num_plugs_used = val(buf$)
  line input #2, buf$
  for i = 1 to 10
    temp$ = FIELD$(buf$, i, ",")
    plugboard_used(i) = val(temp$)
  next i
  line input #2, buf$
  for i = 1 to 10
    temp$ = FIELD$(buf$, i, ",")
    plugboard_used(i+10) = val(temp$)
  next i
  line input #2, buf$
  for i = 1 to 6
    temp$ = FIELD$(buf$, i, ",")
    plugboard_used(i+20) = val(temp$)
  next i
  close #2
end sub

' Convert an RGB color to HSV notation. The RGB components must
' be in the range 0,,255. The S and V values will
' be in range 0..1; the H value will be in range 0..360
sub RGB2HSV r, g, b, h, s, v
  rp = r/255.0 : gp = g/255.0 : bp = b/255.0
  cmax = max(rp, max(gp, bp))
  cmin = min(rp, min(gp, bp))
  delta = cmax - cmin
  if delta = 0 then
    h = 0
  else if cmax = rp then
    h = 60*(((gp-bp)/delta) MOD 6)
  else if cmax = gp then
    h = 60*(((bp-rp)/delta) + 2)
  else
    h = 60*(((rp-gp)/delta) + 4)
  end if
  if cmax = 0 then
    s = 0
  else 
    s = delta/cmax
  end if
  v = cmax
end sub

' Convert an HSV value to its RGB equivalent
' The H and V values must be in range 0..1; the H value must
' be in range 0..360. The RGB values will be in range 0..255.
sub HSV2RGB h, s, v, r, g, b
  hh = h/60.0
  i = int(hh)
  f = hh - i
  p = v*(1-s)
  q = v*(1-s*f)
  t = v*(1-s*(1-f))
  
  select case i
    case 0
      rp = v : gp = t : bp = p
    case 1
      rp = q : gp = v : bp = p
    case 2
      rp = p : gp = v : bp = t
    case 3
      rp = p : gp = q : bp = v
    case 4
      rp = t : gp = p : bp = v
    case 5
      rp = v : gp = p : bp = q
  end select
  r = rp*255 : g = gp*255 : b = bp*255
end sub

' Wiring for Code Wheels I through V and Reflectors b and c
' Each wheel has two lists: the right-to-left list and the
' left-to-right list. Reflectors only have the right-to-left list.
' These data tables cover only the Army models of the Enigma Machine.

' Rotor I
data 5,11,13,6,12,7,4,17,22,26,14,20,15,23,25,8,24,21,19,16,1,9,2,18,3,10
data 21,23,25,7,1,4,6,16,22,26,2,5,3,11,13,20,8,24,19,12,18,9,14,17,15,10

' Rotor II
data 1,10,4,11,19,9,18,21,24,2,12,8,23,20,13,3,17,7,26,14,16,25,6,22,15,5
data 1,10,16,3,26,23,18,12,6,2,4,11,15,20,25,21,17,7,5,14,8,24,13,9,22,19

' Rotor III
data 2,4,6,8,10,12,3,16,18,20,24,22,26,14,25,5,9,23,7,1,11,13,21,19,17,15
data 20,1,7,2,16,3,19,4,17,5,21,6,22,14,26,8,25,9,24,10,23,12,18,11,15,13

' Rotor IV
data 5,19,15,22,16,26,10,1,25,17,21,9,18,8,24,12,14,6,20,7,11,4,3,13,23,2
data 8,26,23,22,1,18,20,14,12,7,21,16,24,17,3,5,10,13,2,19,11,4,25,15,9,6

' Rotor V
data 22,26,2,18,7,9,20,25,21,16,19,4,14,8,12,24,1,23,13,10,17,15,6,5,3,11
data 17,3,25,12,24,23,5,14,6,20,26,15,19,13,22,10,21,4,11,7,9,1,18,16,8,2

' Reflector B
data 25,18,21,8,17,19,12,4,16,24,14,7,15,11,13,9,5,2,6,26,3,23,22,10,1,20

' Reflector C
data 6,22,16,10,9,1,15,25,5,4,18,26,24,23,7,3,20,11,21,17,19,2,14,13,8,12

' Notches for Code Wheels I through V (no notches on reflectors)
data 25, 13, 4, 18, 8

' Turnovers for Code Wheels I through V (no turnovers on reflectors)
data 18, 5, 22, 10, 26

' Keyboard and Lights Ordering (almost but not quite QWERTY)
' Center row only has 8 keys/lights but we add '-' sign to fill array.
data "Q", "W", "E", "R", "T", "Z", "U", "I", "O"
data "A", "S", "D", "F", "G", "H", "J", "K", "-"
data "P", "Y", "X", "C", "V", "B", "N", "M", "L"

' Index for keys, lights, and plugboard showing order of letters
' in raster order. The middle row has a dummy value (0) at end.
data 17, 23, 5, 18, 20, 26, 21, 9, 15
data 1, 19, 4, 6, 7, 8, 10, 11, 0
data 16, 25, 24, 3, 22, 2, 14, 13, 12

#include "Enigma_Help.inc"
